/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/06/22.
//

#ifndef FGGL_ASSETS_MANAGER_HPP
#define FGGL_ASSETS_MANAGER_HPP

#include <string_view>
#include <map>
#include <functional>
#include <memory>
#include <vector>

#include "fggl/debug/logging.hpp"
#include "fggl/assets/types.hpp"
#include "fggl/util/safety.hpp"
#include "fggl/modules/module.hpp"

namespace fggl::assets {

	struct AssetBox {
		virtual ~AssetBox() = default;
		virtual void release() = 0;
	};

	template<typename T>
	struct AssetBoxT : public AssetBox {
		T* asset = nullptr;

		explicit inline AssetBoxT(T* aasset) : asset(aasset) {}

		// no copy: we own our resource!
		AssetBoxT(const AssetBoxT&) = delete;
		AssetBoxT& operator=(const AssetBoxT&) = delete;

		// move OK - we can steal the asset
		AssetBoxT(AssetBoxT&& other) : asset(other.asset) {
			other.asset = nullptr;
		}

		inline ~AssetBoxT() override {
			if ( asset != nullptr ) {
				delete asset;
				asset = nullptr;
			}
		}

		inline void release() override {
			asset = nullptr;
		}
	};

	class AssetManager {
		public:
			constexpr const static modules::ModuleService service = modules::make_service("fggl::assets::Manager");

			AssetManager() = default;
			virtual ~AssetManager() = default;

			// no move, no copy.
			AssetManager(const AssetManager &) = delete;
			AssetManager &operator=(const AssetManager &) = delete;
			AssetManager(AssetManager &&) = delete;
			AssetManager &operator=(AssetManager &&) = delete;

			template<typename T>
			T* get(const AssetID &guid) const {
				try {
					const auto &assetRecord = m_registry.at(guid);
					std::shared_ptr<AssetBoxT<T>> casted = std::dynamic_pointer_cast<AssetBoxT<T>>(assetRecord);
					if ( casted == nullptr ) {
						debug::error("Asset type requested did not match loaded asset type!");
						return nullptr;
					}
					return casted->asset;
				} catch (std::out_of_range& e) {
					return nullptr;
				}
			}

			/**
			 * Pass ownership of the asset to the asset system.
			 *
			 * Once this method is called, the asset system owns the asset, and will free
			 * it when it is no longer required. The asset system assumes it is fully aware
			 * of all usages of the assets it manages via its graph.
			 *
			 * @tparam T the asset type to be managed
			 * @param guid the asset name
			 * @param assetRef the asset itself
			 * @return the owned asset pointer
			 */
			template<typename T>
			T* set(const AssetID &guid, T* assetRef) {
				auto ptr = std::make_shared<AssetBoxT<T>>(assetRef);

				m_registry[guid] = ptr;
				return (*ptr).asset;
			}

			inline void require(const AssetID &/*guid*/) {
				//m_registry.at(guid).refCount++;
			}

			inline bool has(const AssetID &guid) {
				return m_registry.contains(guid);
			}

			inline void release(const AssetGUID &/*guid*/) {
				//m_registry.at(guid).refCount--;
			}

			inline std::vector<AssetID> known() {
				std::vector<AssetID> assetList;
				for ( auto& itr : m_registry ) {
					assetList.push_back(itr.first);
				}
				return assetList;
			}

		private:
			std::map<AssetID, std::shared_ptr<AssetBox>> m_registry;
	};

} // namespace fggl::assets

#endif //FGGL_ASSETS_MANAGER_HPP
